"""Main entry point for the schedule optimizer."""

# %%
import logging
import os
from importlib import metadata
from pathlib import Path

import pandas as pd
import typer
from rich.console import Console
from rich.table import Table
from typing_extensions import Annotated

from optimal_congress.io.api import fetch_events, fetch_rooms
from optimal_congress.io.cache import (
    load_events,
    load_ratings,
    load_rooms,
    save_events,
    save_rating,
    save_rooms,
)
from optimal_congress.optimize import optimize_schedule
from optimal_congress.ratings import (
    enquire_and_save_ratings,
    filter_latest_ratings,
    filter_unrated_events,
    join_events_with_ratings,
)
from optimal_congress.schema import EventLanguage, Rating, RatingsExport, Room

# deactivate color for rich/colorama
os.environ["NO_COLOR"] = "1"

app = typer.Typer(
    add_completion=False,
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=True,
)


# %%
@app.callback()
def users_callback(
    verbose: bool = typer.Option(
        False,
        "-v",
        "--verbose",
        help="Include debug messages in output.",
    ),
) -> None:
    """Optimize your personal schedule for the 38c3."""
    # set log level
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(level=log_level, format="%(message)s")


@app.command()
def fetch(
    dry: bool = typer.Option(
        default=False, help="At dryrun, local cache is not changed."
    ),
) -> None:
    """Fetch events and rooms from API, and update local cache."""

    # fetch from API
    print("Fetching events and rooms from API...")
    events_api = fetch_events()
    rooms_api = fetch_rooms()

    # print summary
    print(f"Fetched {len(events_api)} events and {len(rooms_api)} rooms from API.")

    print("\nComparing API with cache...")
    events_cache = load_events(exit_if_empty=False)

    # check for changes
    new_events = events_api - events_cache
    removed_events = events_cache - events_api
    # report changes
    print(
        f"Found {len(new_events)} new events, and {len(removed_events)} removed events."
    )
    if new_events:
        print("\nNew events:")
        for event in new_events:
            print(f"- {event.name[:50]:.<52}{event.url}")
    if removed_events:
        print("\nRemoved events:")
        for event in removed_events:
            print(f"- {event.name[:50]:.<52}{event.url}")

    # save to cache, if not dryrun
    if dry:
        print("\nDryrun, not updating cache.")
        exit()
    print("\nUpdating cache...")
    save_events(events_api)
    save_rooms(rooms_api)
    print("Done.")


@app.command()
def rate(
    language: str | None = typer.Option(
        None,
        "-l",
        "--language",
        help="Filter events to be rated by language (providing as 2-letter code).",
    ),
) -> None:
    """Interactively rate those events that have not been rated yet.

    Example:
    optimal-congress rate --language en
    """
    # validate input
    if language is not None and language not in EventLanguage.__args__:  # type: ignore
        raise typer.BadParameter(
            f"Invalid language code: {language}. "
            f"Should be one of {list(EventLanguage.__args__)}.",  # type: ignore
        )

    print("loading events and ratings from cache...")
    events = load_events(exit_if_empty=True)
    ratings = load_ratings(exit_if_empty=False)

    print(f"\nFound {len(events)} events and {len(ratings)} ratings.")

    # consider previous ratings
    unrated_events = filter_unrated_events(
        events=events,
        ratings=ratings,
    )
    if len(unrated_events) == 0:
        print("\nNo new events to rate. Exiting.")
        exit()

    # keep only events with relevant language
    if language is not None:
        unrated_events = {
            event
            for event in unrated_events
            if event.language is not None
            if language in event.language
        }
        if len(unrated_events) == 0:
            print(f"\nNo unrated events in chosen language '{language}'. Exiting.")
            exit()

    enquire_and_save_ratings(events=unrated_events)
    print("Done.")


@app.command()
def ratings() -> None:
    """List all latest ratings."""

    print("loading events, ratings, and rooms from cache...")
    events = load_events(exit_if_empty=True)
    ratings = load_ratings(exit_if_empty=True)

    # latest ratings with their events
    latest_ratings = filter_latest_ratings(ratings)
    event_ratings = join_events_with_ratings(
        ratings=latest_ratings,
        events=events,
    )

    # print descenting ratings
    event_ratings_sorted = sorted(
        list(event_ratings),
        key=lambda x: x.rating.score,
        reverse=True,
    )

    # define table
    table = Table(title="Event Ratings")
    table.add_column(header="Rating", justify="right", no_wrap=True)
    table.add_column(header="Title")
    table.add_column(header="URL", justify="center")
    table.add_column(header="Time")

    # populate table
    for event_rating in event_ratings_sorted:
        # format time string
        start_time = event_rating.event.schedule_start.strftime("%a %d %H:%M")
        end_time = event_rating.event.schedule_end.strftime("%H:%M")
        time_string = f"{start_time}-{end_time}"

        table.add_row(
            str(event_rating.rating.score),
            event_rating.event.name[:50],
            f"[link={event_rating.event.url}]🔗[/link]",
            time_string,
        )

    # print table
    print()  # empty line
    console = Console()
    console.print(table)


@app.command()
def optimize(
    minimum_rating: float = typer.Option(
        0.0,
        "-m",
        "--min",
        help="Minimum rating required for talk to be considered in optimization.",
    ),
) -> None:
    """Optimize the schedule based on ratings."""

    print("loading events, ratings, and rooms from cache...")
    ratings = load_ratings(exit_if_empty=True)
    events = load_events(exit_if_empty=True)
    rooms: set[Room] = load_rooms(exit_if_empty=True)

    # latest ratings with their events
    latest_ratings = filter_latest_ratings(ratings)
    event_ratings = join_events_with_ratings(
        ratings=latest_ratings,
        events=events,
    )

    # filter events by minimum required rating
    event_ratings_filtered = {
        event_rating
        for event_rating in event_ratings
        if event_rating.rating.score >= minimum_rating
    }

    # optimize schedule
    scheduled_events = optimize_schedule(event_ratings=event_ratings_filtered)

    events_sorted = sorted(
        scheduled_events, key=lambda event: event.schedule_start, reverse=False
    )

    # define table
    table = Table(title="\nScheduled events:")
    table.add_column(header="Time")
    table.add_column(header="Room")
    table.add_column(header="Title")
    table.add_column(header="URL", justify="center")

    # populate table
    for event in events_sorted:
        # get room name via event's room id
        try:
            room_name = {room for room in rooms if room.id == event.room}.pop().name
        except KeyError:
            room_name = str()

        # format time string
        start_time = event.schedule_start.strftime("%a %d %H:%M")
        end_time = event.schedule_end.strftime("%H:%M")
        time_string = f"{start_time} - {end_time}"

        table.add_row(
            time_string,
            room_name,
            event.name[:60],
            f"[link={event.url}]🔗[/link]",
        )

    # print table
    print()  # empty line
    console = Console()
    console.print(table)


@app.command()
def next(
    min_rating: float = typer.Option(
        5.0,
        "-r",
        "--rating",
        help="Minimum rating required for event to be listed.",
    ),
    num_events: int = typer.Option(
        10,
        "-n",
        "--number",
        help="Number of events to list.",
    ),
) -> None:
    """List next upcoming events, filtered by minimum rating."""
    # get current time
    now = pd.Timestamp.now(tz="Europe/Berlin")

    print("loading events, ratings, and rooms from cache...")
    events = load_events(exit_if_empty=True)
    rooms: set[Room] = load_rooms(exit_if_empty=True)

    # join ratings with their events
    event_ratings = join_events_with_ratings(
        ratings=load_ratings(exit_if_empty=True),
        events=events,
    )

    # filter events by minimum required rating, and 'is in future'
    events_filtered = {
        event_rating.event
        for event_rating in event_ratings
        if event_rating.rating.score >= min_rating
        and event_rating.event.schedule_start > now
    }

    # sort events by start time, keep only first n
    events_sorted = sorted(
        events_filtered, key=lambda event: event.schedule_start, reverse=False
    )[:num_events]

    # print scheduled events
    print("\nNext events:")
    for event in events_sorted:
        # get room name via event's room id
        try:
            room_name = {room for room in rooms if room.id == event.room}.pop().name
        except KeyError:
            room_name = str()

        start_time = event.schedule_start.strftime("%a %d %H:%M")
        end_time = event.schedule_end.strftime("%H:%M")
        print(
            f"- {start_time}-{end_time} {room_name[:15]:.<16}"
            f"{event.name[:40]:.<42}{event.url}"
        )


@app.command()
def dump(
    file_path: Annotated[
        str,
        typer.Argument(
            help="Relative or absolute path to which CSV file will be exported.",
        ),
    ],
) -> None:
    """Export all latest ratings to CSV, for bulk editing.

    This will exports the latest rating for each rated event.
    """
    # convert argument to absolute path
    path = Path(file_path)
    absolute_path = path if path.is_absolute() else Path.cwd() / path

    # load events and ratings
    print("loading events and ratings from cache...")
    events = load_events(exit_if_empty=True)
    ratings = load_ratings(exit_if_empty=True)

    # latest ratings with their events
    latest_ratings = filter_latest_ratings(ratings)
    event_ratings = join_events_with_ratings(
        ratings=latest_ratings,
        events=events,
    )

    # export to CSV
    print(f"Exporting {len(event_ratings)} ratings to {absolute_path}...")
    ratings_df = RatingsExport(
        pd.DataFrame(
            data=[
                [
                    event_rating.rating.score,
                    event_rating.event.name,
                    event_rating.event.url,
                    event_rating.event.id,
                ]
                for event_rating in event_ratings
            ],
            columns=["rating", "name", "url", "event_id"],
        ).sort_values(by="rating", ascending=False)
    )

    ratings_df.to_csv(path_or_buf=absolute_path, index=False)
    print("Done.")


@app.command()
def load(
    file_path: Annotated[
        str,
        typer.Argument(
            help="Name and relative or absolute path from which to read in CSV.",
        ),
    ],
    dry: bool = typer.Option(
        default=False, help="At dryrun, local cache is not changed."
    ),
) -> None:
    """Bulk import ratings from CSV.

    This will overwrite existing ratings.

    CSV format should be the same one as what the `dump` command exports.
    """
    # convert argument to absolute path
    path = Path(file_path)
    absolute_path = path if path.is_absolute() else Path.cwd() / path

    # load ratings from CSV
    ratings_df = RatingsExport(
        pd.read_csv(filepath_or_buffer=absolute_path, index_col=False)
    )
    if len(ratings_df) == 0:
        print("No ratings found in CSV. Exiting.")
        exit()

    # convert to Rating objects
    ratings = {
        Rating(
            event_id=row["event_id"],
            score=row["rating"],
        )
        for _, row in ratings_df.iterrows()
    }

    # save ratings
    if dry:
        print("\nDryrun, not updating cache.")
        exit()
    print(f"Saving {len(ratings)} ratings to cache...")
    for rating in ratings:
        save_rating(rating=rating)


@app.command()
def version() -> None:
    """Print version and exit."""
    version = metadata.version(distribution_name="optimal-congress")
    print(version)


# %%
if __name__ == "__main__":
    app()



"""Schedule optimization."""

import logging

from pulp import PULP_CBC_CMD, LpMaximize, LpProblem, LpStatus, LpVariable

from optimal_congress.schema import Event, EventRating, events_overlap


def optimize_schedule(
    event_ratings: set[EventRating],
) -> set[Event]:
    """
    Optimize the schedule of events based on ratings.

    Args:
        events_ratings: Tuples of events and matching ratings.
    Returns:
        Scheduled events.
    Raises:
        ValueError: If no optimal solution is found.
    """
    # unpack events and ratings
    events = [event_rating.event for event_rating in event_ratings]
    ratings = [event_rating.rating for event_rating in event_ratings]

    # sanity check
    assert len(events) == len(ratings)

    # define problem
    prob = LpProblem(name="OptimalCongress", sense=LpMaximize)

    # define decision variables (binary vector of same length as events)
    lp_vars = [
        LpVariable(
            name=event.slug,
            cat="Binary",
        )
        for event in events
    ]

    # objective function: maximize sum of ratings for scheduled events
    prob += sum(lp_var * rating.score for lp_var, rating in zip(lp_vars, ratings))

    # constraints: no overlapping events can be scheduled
    for i, event_i in enumerate(events):
        for j, event_j in enumerate(events):
            if i >= j:
                # avoid double counting, and comparing event to itself
                continue
            if events_overlap(event_i, event_j):
                # rename event names to make work with pulp
                event_i_name = event_i.slug.replace("-", "_")
                event_j_name = event_j.slug.replace("-", "_")
                constraint_name = f"overlap_{event_i_name}_{event_j_name}"
                prob += (lp_vars[i] + lp_vars[j] <= 1, constraint_name)

    logging.debug("\nProblem:")
    logging.debug(prob)

    # solve problem
    prob.solve(PULP_CBC_CMD(msg=False))

    # check if optimal solution was found
    optimal = LpStatus[prob.status] == "Optimal"
    if not optimal:
        raise ValueError("No optimal solution found.")

    logging.debug("solution:")
    for var in prob.variables():
        logging.debug(f"{var.name}: {var.varValue}")

    # extract scheduled events
    scheduled_event_names: set[str] = {
        var.name for var in prob.variables() if var.varValue == 1
    }
    scheduled_events: set[Event] = {
        event
        for event in events
        if event.slug.replace("-", "_") in scheduled_event_names
    }
    return scheduled_events



"""IO operations on local cache."""

import json

from optimal_congress.config import (
    DIR_EVENTS_CACHE,
    DIR_RATINGS_CACHE,
    DIR_ROOMS_CACHE,
)
from optimal_congress.schema import Event, Rating, Room


def save_events(
    events: set[Event],
    clear: bool = True,
) -> None:
    """Save events to cache.

    Args:
        events: List of events to save.
        clear: Whether to clear all cached events before saving. Defaults to True.
    """
    # create events cache directory, if it does not exist
    DIR_EVENTS_CACHE.mkdir(parents=True, exist_ok=True)

    # empty cache
    if clear:
        for file in DIR_EVENTS_CACHE.glob("*.json"):
            file.unlink()

    # save events
    for event in events:
        with open(DIR_EVENTS_CACHE / f"event_{event.id}.json", "w") as f:
            f.write(event.model_dump_json())


def save_rooms(
    rooms: set[Room],
    clear: bool = True,
) -> None:
    """Save rooms to cache.

    Args:s
        rooms: List of rooms to save.
        clear: Whether to clear all cached rooms before saving. Defaults to True.
    """
    # create rooms directory if it doesn't exist
    DIR_ROOMS_CACHE.mkdir(parents=True, exist_ok=True)

    # empty cache
    if clear:
        for file in DIR_ROOMS_CACHE.glob("*.json"):
            file.unlink()

    # save rooms
    for room in rooms:
        with open(DIR_ROOMS_CACHE / f"room_{room.id}.json", "w") as f:
            f.write(room.model_dump_json())


def save_rating(rating: Rating) -> None:
    """Save single new rating to cache.

    Note: This does not overwrite cached rating for the same event.
    """
    # create ratings directory if it doesn't exist
    DIR_RATINGS_CACHE.mkdir(parents=True, exist_ok=True)

    # save ratings
    with open(
        DIR_RATINGS_CACHE / f"rating_{rating.event_id}_{rating.timestamp}.json",
        "w",
    ) as f:
        f.write(rating.model_dump_json())


def load_events(exit_if_empty: bool) -> set[Event]:
    """Load events from disk.

    Args:
        exit_if_empty: Exit if no events are found, and give instructions.
    Returns:
        List of events.
    """
    # create events directory if it doesn't exist
    DIR_EVENTS_CACHE.mkdir(parents=True, exist_ok=True)

    # load events
    events_files = list(DIR_EVENTS_CACHE.glob("*.json"))
    events = {Event(**json.loads(open(file).read())) for file in events_files}

    # exit if no events are found
    if exit_if_empty and len(events) == 0:
        print("\nNo events found! Run `fetch` command to load events from API.")
        exit()
    return events


def load_rooms(exit_if_empty: bool) -> set[Room]:
    """Load rooms from disk.

    Args:
        exit_if_empty: Exit if no events are found, and give instructions.
    Returns:
        List of rooms.
    """
    # create rooms directory if it doesn't exist
    DIR_ROOMS_CACHE.mkdir(parents=True, exist_ok=True)

    # load rooms
    rooms_files = list(DIR_ROOMS_CACHE.glob("*.json"))
    rooms = {Room(**json.loads(open(file).read())) for file in rooms_files}

    # exit if no events are found
    if exit_if_empty and len(rooms) == 0:
        print("\nNo rooms found! Run `fetch` command to load room info from API.")
        exit()
    return rooms


def load_ratings(exit_if_empty: bool) -> set[Rating]:
    """Load all ratings from disk.

    Args:
        exit_if_empty: Exit if no ratings are found, and give instructions.
    Returns:
        List of ratings.
    """
    # create ratings directory if it doesn't exist
    DIR_RATINGS_CACHE.mkdir(parents=True, exist_ok=True)

    # load ratings
    ratings_files = list(DIR_RATINGS_CACHE.glob("*.json"))
    ratings = {Rating(**json.loads(open(file).read())) for file in ratings_files}

    # exit if no events are found
    if exit_if_empty and len(ratings) == 0:
        print("\nNo ratings found! Run `rate` command to rate events.")
        exit()
    return ratings



"""Model definitions."""

from datetime import datetime
from typing import Annotated
from uuid import UUID

import pandera as pa
from pydantic import BaseModel, BeforeValidator, Field
from typing_extensions import Literal

from optimal_congress.config import HUB_EVENT_ROUTE

# languages to accept for events
EventLanguage = Literal["de", "en"]


class Room(BaseModel):
    """A room."""

    id: UUID
    name: str
    assembly: str

    class Config:
        frozen = True  # instances immutable and hashable


def parse_language(
    value: list[EventLanguage] | str | None,
) -> list[EventLanguage] | None:
    """Parse a language value to the expected list of languages.

    Args:
        value: The language value to parse.
    Returns:
        The parsed list of languages, or None.
    """
    match value:
        case None:
            return None
        case list():
            return value
        case str():
            return [i.strip() for i in value.split(",")]  # type: ignore


class Event(BaseModel):
    """An event."""

    id: UUID
    name: str
    slug: str
    track: str | None
    assembly: str
    room: UUID | None
    language: Annotated[
        list[EventLanguage] | None,
        BeforeValidator(parse_language),
    ] = Field(default=None)
    description: str
    schedule_start: datetime
    schedule_end: datetime

    def __hash__(self) -> int:
        """Events are equal, if they have same ID - regardless of other properties."""
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        """Events are equal, if they have same ID - regardless of other properties."""
        if not isinstance(other, Event):
            return NotImplemented
        return self.id == other.id

    def __str__(self) -> str:
        """Return a string representation of the event."""
        # get time without timezone
        timestamp_start = self.schedule_start.strftime("%Y-%m-%d %H:%M")
        time_end = self.schedule_end.strftime("%H:%M")
        return f"'{self.name}' ({timestamp_start} - {time_end}, {self.url})"

    @property
    def url(self) -> str:
        """Return the url of the event."""
        return f"{HUB_EVENT_ROUTE}/{self.slug}"


def events_overlap(event1: Event, event2: Event) -> bool:
    """Check if two events overlap."""
    return (
        event1.schedule_start < event2.schedule_end
        and event2.schedule_start < event1.schedule_end
    )


class Rating(BaseModel):
    """A rating for an event."""

    event_id: UUID
    score: float = Field(description="A positive number describing utility to attend.")
    timestamp: datetime = Field(default_factory=datetime.now)

    class Config:
        frozen = True  # instances immutable and hashable


class EventRating(BaseModel):
    """A container class for a rating and its associated event."""

    event: Event
    rating: Rating

    class Config:
        frozen = True  # instances immutable and hashable


class RatingsExport(pa.typing.DataFrame):
    """A schema for exporting and importing ratings to/from CSV."""

    score: float = pa.Field()
    name: str = pa.Field()
    url: str = pa.Field()
    id: UUID = pa.Field()



"""App configuration."""

from pathlib import Path

# URLS
API_EVENTS = "https://api.events.ccc.de/congress/2024/events"
API_ROOMS = "https://api.events.ccc.de/congress/2024/rooms"
HUB_EVENT_ROUTE = "https://events.ccc.de/congress/2024/hub/en/event"

# folder to store serialized ratings
DIR_RATINGS_CACHE = Path.home() / ".cache/congress_optimizer/ratings"
DIR_EVENTS_CACHE = Path.home() / ".cache/congress_optimizer/events"
DIR_ROOMS_CACHE = Path.home() / ".cache/congress_optimizer/rooms"



"""Functions related to ratings."""

import os

from optimal_congress.io.cache import save_rating
from optimal_congress.schema import Event, EventRating, Rating


def filter_latest_ratings(ratings: set[Rating]) -> set[Rating]:
    """Return the latest rating for each event.

    Args:
        ratings: List of ratings.
    Returns:
        List of latest ratings.
    """
    # sort ratings by timestamp, descending
    ratings_sorted = sorted(
        list(ratings), key=lambda rating: rating.timestamp, reverse=True
    )

    # keep only latest rating for each event
    latest_ratings: set[Rating] = set()
    for rating in ratings_sorted:
        if rating.event_id not in [r.event_id for r in latest_ratings]:
            latest_ratings.add(rating)
    return latest_ratings


def filter_unrated_events(
    events: set[Event],
    ratings: set[Rating],
) -> set[Event]:
    """Return the unrated events from a list of events based on previous ratings.

    Args:
        events: list of events.
        previous_ratings: list of previous ratings.
    Returns:
        List of events for which no rating is provided.
    """
    rated_event_ids = {rating.event_id for rating in ratings}
    unrated_events = {event for event in events if event.id not in rated_event_ids}
    return unrated_events


def enquire_and_save_ratings(events: set[Event]) -> None:
    """Enquire and save ratings for a list of events."""

    for i, event in enumerate(events):
        os.system("cls" if os.name == "nt" else "clear")
        timestamp_start = event.schedule_start.strftime("%Y-%m-%d %H:%M")
        time_end = event.schedule_end.strftime("%H:%M")
        print(
            f"\nUnrated event ({i + 1}/{len(events)}):"
            f"\n\n{timestamp_start} - {time_end}"
            f"\n\n{event.name}"
            f"\n\n{event.description}"
            f"\n\n{event.url}"
        )
        try:
            score = input("\nRate from 0 to 10 (Enter to exit): ")
            if score == "":
                raise KeyboardInterrupt
        except KeyboardInterrupt:
            print("\nExiting.")
            break

        rating = Rating(event_id=event.id, score=float(score))

        # save rating
        print(f"Saving rating '{rating.score}' for event '{event.name}'...")
        save_rating(rating=rating)


def join_events_with_ratings(
    ratings: set[Rating],
    events: set[Event],
) -> set[EventRating]:
    """Inner-Join events with ratings.

    Note that duplicated ratings are not removed, if they are provided as input.

    Args:
        events: Set of events.
        ratings: Set of ratings.
    Returns:
        Ratings with their associated Events.
    """
    # join events with ratings
    event_ratings: set[EventRating] = set()
    for rating in ratings:
        matching_events = {event for event in events if event.id == rating.event_id}
        if matching_events:
            event = {event for event in events if event.id == rating.event_id}.pop()
            event_ratings.add(EventRating(event=event, rating=rating))

    return event_ratings



"""Functions for read operations against congress API."""

import json

import requests

from optimal_congress.config import API_EVENTS, API_ROOMS
from optimal_congress.schema import Event, Room


def fetch_events() -> set[Event]:
    """Load events from the API."""
    events_dict = json.loads(s=requests.get(url=API_EVENTS).text)
    events = {Event(**event) for event in events_dict}
    if len(events) == 0:
        raise ValueError("No events found! Check state of congress API.")
    return events


def fetch_rooms() -> set[Room]:
    """Load rooms from the API."""
    rooms_dict = json.loads(s=requests.get(url=API_ROOMS).text)
    rooms = {Room(**room) for room in rooms_dict}
    if len(rooms) == 0:
        raise ValueError("No rooms found! Check state of congress API.")
    return rooms
